import { isObjectProperty, isIdentifier, isStringLiteral, binaryExpression, stringLiteral, templateLiteral, templateElement, objectProperty, identifier, booleanLiteral, objectExpression, arrayExpression, callExpression } from '@babel/types';
import { registerPreset, transform } from '@babel/standalone';
import register, { revert } from '@babel/register';
import evaluatePath from 'babel-helper-evaluate-path';
import resolve from 'resolve';
import path from 'path';
import * as taddy from 'taddy';
import { $css, ruleInjector } from 'taddy';
import tsPreset from '@babel/preset-typescript';
import reactPreset from '@babel/preset-react';
import envPreset from '@babel/preset-env';
import { VARS_KEY, config } from '@taddy/core';
import fs from 'fs';

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;

  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }

  return target;
}

/* eslint-disable */
const MACRO_NAME = 'taddy.macro';
const PACKAGE_NAME = 'taddy'; // TODO: add config resolution

/*
function loadConfig(filepath: string): object {
    const {Project} = require('ts-morph');

    // empty project to parse the config
    const project: TSProject = new Project();

    const sourceFile = project.addSourceFileAtPath(filepath);

    const decl = sourceFile.getDefaultExportSymbol();

    const properties = getType(decl)
        ?.getProperties()
        .find((x) => x.getEscapedName() === 'properties');

    const result = parseObject(getType(properties));

    const compiled = project.emitToMemory().getFiles()[0].text;

    // eslint-disable-next-line
    const emit = new Function('exports', `${compiled};return exports;`);

    return {
        ...emit({}).default,
        _properties: result,
    };
}

export const config = lilconfigSync(PACKAGE_NAME, {
    searchPlaces: [`${PACKAGE_NAME}.config.ts`],
    loaders: {
        '.ts': loadConfig,
    },
}).search()?.config;
*/

function getRootDir() {
  return process.cwd();
}
function getCacheDir() {
  return __dirname;
}

const cache = /*#__PURE__*/new WeakMap();
class BindingError extends Error {}

function addBinding(bindingMap, binding, path) {
  bindingMap.set(binding, (bindingMap.get(binding) || new Set()).add(path));
}

function isFunctionArgument(path) {
  let bottomup = 3;

  while (bottomup >= 0) {
    if (path.listKey === 'params' && path.parentPath.isFunction()) {
      return true;
    }

    bottomup--;
  }

  return false;
}

function findBindings(currentPath) {
  function visitIdentifier(referentPath, path, bindingMap) {
    if (path.parentPath.isMemberExpression()) {
      // @ts-expect-error
      path = path.parentPath.get('object');
    } else if (path.parentPath.isObjectPattern()) {
      // @ts-expect-error
      path = path.parentPath.parentPath;
    }

    const {
      parentPath,
      node
    } = path;

    if (!(parentPath.isExpression() || parentPath.isSpreadElement() || parentPath.isObjectProperty() && (parentPath.node.value === node || parentPath.node.computed))) {
      return;
    }

    if (!('name' in node && typeof node.name === 'string')) return;
    const {
      name
    } = node;
    const sharedBinding = currentPath.scope.getBinding(name);
    const binding = referentPath.scope.getBinding(name);
    if (!binding) return;
    if (sharedBinding !== binding) return;

    if (binding.constantViolations.length > 0) {
      throw new BindingError('CONSTANT VIOLATION');
    }

    const {
      path: bindingPath
    } = binding;

    if (bindingMap.has(binding)) {
      addBinding(bindingMap, binding, path);
      return;
    }

    if (isFunctionArgument(bindingPath)) {
      throw new BindingError('FUNCTION ARGUMENT');
    } // addBinding(bindingMap, binding, path);


    if (!(bindingPath.listKey === 'params')) {
      addBinding(bindingMap, binding, path);
    }

    if (bindingPath.isImportSpecifier()) {
      return;
    }

    for (let [newBinding, paths] of traverse(bindingPath)) {
      paths.forEach(x => {
        addBinding(bindingMap, newBinding, x);
      });
    }
  }

  const visited = new Set();

  function traverse(referentPath) {
    if (referentPath.isObjectPattern()) ;

    if (cache.has(referentPath)) {
      return cache.get(referentPath);
    }

    const bindings = new Map(); // we should not revisit the paths, otherwise something went wrong

    if (visited.has(referentPath)) {
      return bindings;
    }

    visited.add(referentPath);

    if (referentPath.isIdentifier()) {
      visitIdentifier(referentPath, referentPath, bindings);
    }

    referentPath.traverse({
      Identifier(path) {
        visitIdentifier(referentPath, path, bindings);
      }

    });
    cache.set(referentPath, bindings);
    return bindings;
  }

  return traverse(currentPath);
}

function getPathsByBindings(bindings) {
  const paths = new Set();

  for (let x of bindings.keys()) {
    if (x.path.isImportSpecifier() || x.path.isVariableDeclarator()) {
      paths.add(x.path.parentPath);
      continue;
    }

    paths.add(x.path);
  }

  return paths;
}

function buildCodeByPath(path) {
  return Array.from(getPathsByBindings(findBindings(path))).sort((a, b) => a.node.start - b.node.start).map(x => x.toString()).join('\n');
}

registerPreset('@babel/preset-typescript', tsPreset);
registerPreset('@babel/preset-react', reactPreset);
registerPreset('@babel/preset-env', envPreset);
const DEFAULT_PRESETS = [['@babel/preset-typescript', {
  allExtensions: true,
  isTSX: true
}], '@babel/preset-react', ['@babel/preset-env', {
  targets: {
    node: 'current'
  }
}]];
const EXTENSIONS = ['.es6', '.es', '.tsx', '.ts', '.jsx', '.js', '.mjs'];
const macroRe = /*#__PURE__*/new RegExp( /*#__PURE__*/MACRO_NAME.replace('.', '\\.'), 'g');
const EVAL_FILENAME_POSTFIX = '@__TADDY_EVALUATE__'; // webpack "require" critical dependency issue workaround

const nodeRequire = /*#__PURE__*/new Function('require', `return typeof require !== 'undefined'
        ? require
        : (typeof globalThis !== 'undefined' ? globalThis : global).require; `)(module.require);
function isTaddyEvaluation(state) {
  return state.filename.includes(EVAL_FILENAME_POSTFIX);
}
function evaluate(currentPath) {
  const result = evaluatePath(currentPath);

  if (result.confident) {
    return {
      value: result.value
    };
  }

  let content = '';
  let code = '';

  try {
    const callbackName = '__taddy__';
    content = buildCodeByPath(currentPath) // hack to avoid processing by babel-macro
    .replace(macroRe, PACKAGE_NAME).concat(`\n\n;${callbackName}(${currentPath.toString()})`);
    const {
      opts
    } = currentPath.hub.file;
    const ext = path.extname(opts.filename);
    const basename = path.basename(opts.filename, ext);
    const dirname = path.dirname(opts.filename);
    const options = {
      babelrc: false,
      configFile: false,
      filename: path.join(dirname, basename + EVAL_FILENAME_POSTFIX + ext),
      plugins: [
        /*...opts.plugins*/
      ],
      presets: [
      /*...opts.presets*/
      ...DEFAULT_PRESETS]
    };
    ({
      code
    } = transform(content, options) || {}); // console.log('evaluate', {code});

    if (!code) return {};
    const exec = new Function('require', callbackName, code);
    register(_extends({}, options, {
      extensions: EXTENSIONS
    }));
    let value;
    exec(filepath => {
      if (filepath === PACKAGE_NAME) return taddy;
      const requirePath = resolve.sync(filepath, {
        extensions: EXTENSIONS,
        basedir: path.dirname(opts.filename)
      });
      return nodeRequire(requirePath);
    }, result => {
      value = result;
    });
    revert();
    return {
      value
    };
  } catch (error) {
    // console.log('evaluate error', {content, code, error});
    return {
      error
    };
  }
}

function getObjectPropertyKey(path) {
  if (path.isIdentifier()) {
    return path.node.name;
  }

  if (path.isLiteral() && 'value' in path.node) {
    return String(path.node.value);
  }

  return null;
}

function mergeObjectProperties(properties) {
  const map = new Map();

  function move(key, value) {
    map.delete(key);
    map.set(key, value);
  }

  for (const x of properties) {
    if (!isObjectProperty(x)) {
      move(x, x);
      continue;
    }

    const {
      key
    } = x;

    if (isIdentifier(key)) {
      move(key.name, x);
      continue;
    }

    if (isStringLiteral(key)) {
      move(key.value, x);
      continue;
    }

    move(x, x);
  } // TODO: there was a problem with typescript compiler with the variant [...map.values()]


  return Array.from(map.values());
}
function mergeObjects(objects) {
  const properties = [];
  objects.forEach(obj => {
    properties.push(...obj.properties);
  });
  return mergeObjectProperties(properties);
}

function isTaddy(binding) {
  const {
    path
  } = binding;
  if (!path.isImportSpecifier()) return false;
  return path.parentPath.node.source.value === 'taddy' && path.node.imported.name === 'css' || path.node.imported.name === 'mixin';
}

function optimizeBindings(referentPath, {
  env
} = {}) {
  let bindings;

  try {
    bindings = findBindings(referentPath);
  } catch (error) {
    return;
  } // console.log(bindings);


  for (let [binding] of bindings) {
    if (isTaddy(binding)) continue;
    optimizeBindings(binding.path);
  }

  for (let [binding, paths] of bindings) {
    if (isTaddy(binding)) continue;
    if (binding.path.removed) continue;

    if (binding.references > paths.size) {
      continue;
    }

    if (binding.references < paths.size) {
      // that should not be possible, but if so,
      // then we can't be sure that it's correct to remove this binding
      // TODO: think if we can handle this somehow
      continue;
    }

    const {
      parentPath
    } = binding.path;
    const isImportToRemove = parentPath.isImportDeclaration() && parentPath.node.specifiers.length === 0; // // keep imports for development build
    // if (env === 'development') {
    //     return;
    // }

    binding.path.remove();

    if (!isImportToRemove) {
      continue;
    }

    parentPath.remove();
  }
}

function optimizeStaticStyles(path) {
  const quasis = [];
  const expressions = [];
  const props = [];

  for (const propPath of path.get('properties')) {
    if (!propPath.isObjectProperty()) {
      throw new Error('Cant optimize this path');
    }

    const keyPath = propPath.get('key');
    const key = getObjectPropertyKey(keyPath);
    if (!key) return;
    const valuePath = propPath.get('value');

    if (key === 'className') {
      expressions.push(valuePath.node);
      continue;
    }

    if (key === 'style' || key === VARS_KEY) {
      props.push(propPath.node);
      continue;
    }

    if (valuePath.isStringLiteral()) {
      quasis.push(key + valuePath.node.value);
      continue;
    }

    if (valuePath.isBooleanLiteral()) {
      quasis.push(key);
      continue;
    }

    expressions.push(binaryExpression('+', stringLiteral(key), valuePath.node));
  }

  let classes = stringLiteral(quasis.join(' '));

  if (expressions.length !== 0) {
    classes.value += classes.value ? ' ' : '';
    const left = classes;
    const right = expressions.length === 1 ? expressions[0] : templateLiteral([templateElement({
      raw: ''
    })].concat([...Array(expressions.length - 1)].fill(templateElement({
      raw: ' '
    }))).concat(templateElement({
      raw: ''
    })), expressions);
    classes = binaryExpression('+', left, right);
  }

  if (props.length === 0) {
    path.replaceWith(classes);
    return;
  }

  path.node.properties = props.concat(objectProperty(identifier('className'), classes));
}

function getType(symbol) {
  return symbol === null || symbol === void 0 ? void 0 : symbol.getTypeAtLocation(symbol.getDeclarations()[0]);
}
class DynamicType {
  constructor({
    value,
    type
  }) {
    this.value = value;
    this.type = type;
  }

}
class UnionType extends Array {}

function isRecord(type) {
  var _type$getAliasSymbol;

  return (type === null || type === void 0 ? void 0 : (_type$getAliasSymbol = type.getAliasSymbol()) === null || _type$getAliasSymbol === void 0 ? void 0 : _type$getAliasSymbol.getEscapedName()) === 'Record';
}

const STATIC_KEY = /*#__PURE__*/Symbol('__TADDY_STATIC_KEY__');
function isStaticValue(value) {
  if (!value || typeof value !== 'object') {
    return true;
  }

  if (value instanceof UnionType) {
    return false;
  }

  return value[STATIC_KEY] || false;
}

function filterValue(value) {
  // filter vendor prefixes
  return String(value).startsWith('-');
}

function parseValue(type) {
  var _value;

  if (!type) return null;

  if (type.isAny() || type.isUnknown()) {
    return new DynamicType({
      value: null,
      type
    });
  }

  if (isRecord(type)) {
    return parseRecord(type);
  }

  if (type.isArray()) {
    return parseArray(type);
  }

  if (type.isUnion()) {
    return new UnionType(...type.getUnionTypes().map(parseValue));
  }

  if (type.isObject()) {
    return parseObject(type);
  }

  let value = type === null || type === void 0 ? void 0 : type.getText();
  /**
   * Firstly, exclude types like "string & {}"
   *
   * TODO: think about other intersection types
   */

  if (!value || ((_value = value) === null || _value === void 0 ? void 0 : _value.includes('&'))) {
    return null;
  }

  try {
    value = JSON.parse(value);
  } catch (e) {
    return new DynamicType({
      value,
      type
    });
  }

  if (filterValue(value)) return null;
  return value;
}

function parseMixinParameters(params) {
  const type = getType(params);
  return parseValue(type);
}

function parseMixin(mixin) {
  const type = getType(mixin);

  if ((type === null || type === void 0 ? void 0 : type.getCallSignatures().length) === 0) {
    return parseMixinParameters(mixin);
  }

  return type === null || type === void 0 ? void 0 : type.getCallSignatures()[0].getParameters().map(parseMixinParameters);
}

function parseRecord(type) {
  const [k, v] = type.getAliasTypeArguments();
  const key = parseValue(k);
  const value = parseValue(v);

  if (!Array.isArray(key)) {
    return {
      [key]: value,
      [STATIC_KEY]: value[STATIC_KEY]
    };
  }

  return key.map(prop => ({
    [prop]: value,
    [STATIC_KEY]: value[STATIC_KEY]
  }));
} // TODO: improve the array parsing (such as not just union types)


function parseArray(type) {
  const elements = [];
  elements[STATIC_KEY] = true;

  function applyElement(element) {
    const value = parseValue(element);
    if (!value) return;
    const isStatic = isStaticValue(value);
    elements[STATIC_KEY] = elements[STATIC_KEY] && isStatic;
    elements.push(isStatic ? value : new Set([].concat(value).filter(Boolean)));
  }

  const elementType = type.getArrayElementTypeOrThrow();

  for (let element of elementType.getUnionTypes() || [elementType]) {
    applyElement(element);
  }

  return elements;
}

function parseObject(type) {
  const props = {};
  let isStaticObject = true;

  for (const propSymbol of (_type$getProperties = type === null || type === void 0 ? void 0 : type.getProperties()) !== null && _type$getProperties !== void 0 ? _type$getProperties : []) {
    var _type$getProperties;

    const name = propSymbol.getEscapedName();
    const value = parseMixin(propSymbol);
    if (!value) continue;
    const isStatic = isStaticValue(value);
    isStaticObject = isStaticObject && isStatic;
    props[name] = isStatic ? value : new Set([].concat(value).filter(Boolean));
  }

  props[STATIC_KEY] = isStaticObject;
  return props;
}
class TSProcessor {
  constructor(config) {
    const {
      Project
    } = require('ts-morph');

    if (config instanceof Project) {
      this.project = config;
      return;
    }

    this.project = new Project(_extends({
      tsConfigFilePath: path.join(process.cwd(), 'tsconfig.json')
    }, config));
  }

  getTypeAtPos(filename, code, start, end) {
    let sourceFile = this.project.getSourceFile(filename);

    if (sourceFile && sourceFile.getFullText() !== code) {
      sourceFile.replaceText([0, sourceFile.getEnd()], code);
    } else {
      sourceFile = this.project.createSourceFile(filename, code, {
        overwrite: true
      });
    }

    const typeNode = sourceFile.getDescendantAtStartWithWidth(start, end - start);
    return typeNode;
  }

}

function getLiteralValue(path) {
  try {
    // eslint-disable-next-line no-eval
    return eval(path.toString());
  } catch (error) {
    return getLiteralValue.FAIL;
  }
}

getLiteralValue.FAIL = /*#__PURE__*/Symbol('FAILED_VALUE');

function getHashedName(key, {
  postfix
}) {
  return config.nameGenerator.getName(key, '', {
    postfix
  }).join('');
}

class Processor {
  constructor({
    config
  }) {
    this.config = config;

    if (config.typescript) {
      this.tsProject = new TSProcessor(config.typescript === true ? {} : config.typescript);
    }
  }

  prepare({
    mixin
  }) {
    this.mixin = mixin;
    this.isStatic = false;
    this.variables = [];
    this.classNameNodes = new WeakSet();
    this.optimizationPaths = new Set();
  }

  isClassNameNode(node) {
    return this.classNameNodes.has(node);
  }

  classNamesToNode(classNames) {
    const properties = [];

    for (const key in classNames) {
      const classValue = classNames[key];
      const propName = identifier(key);
      const propValue = typeof classNames[key] === 'boolean' ? booleanLiteral(classValue) : stringLiteral(String(classValue));
      const node = objectProperty(propName, propValue);
      this.classNameNodes.add(node);
      properties.push(node);
    }

    return properties;
  }

  stylesToNode(key, value, {
    postfix = ''
  } = {}) {
    const {
      className
    } = $css({
      [key]: value
    }, {
      postfix
    });
    return this.classNamesToNode(className);
  }

  processPropertyValue(key, path, {
    postfix,
    properties
  }) {
    const keyPath = path.get('key');
    const valuePath = path.get('value');

    const tryLiteralValue = () => {
      if (!valuePath.isLiteral()) return false;
      const value = getLiteralValue(valuePath);

      if (value === getLiteralValue.FAIL) {
        return false;
      }

      properties.push(...this.stylesToNode(key, value, {
        postfix
      }));
      return true;
    };

    const tryObjectExpressionValue = () => {
      if (!valuePath.isObjectExpression()) return false;
      this.process(valuePath, {
        postfix: postfix + key
      });
      let currentStyles = [];

      for (const nestedProp of valuePath.node.properties) {
        if (this.isClassNameNode(nestedProp)) {
          if (currentStyles.length) {
            properties.push(objectProperty(keyPath.node, objectExpression(currentStyles)));
            currentStyles = [];
          }

          properties.push(nestedProp);
        } else {
          currentStyles.push(nestedProp);
        }
      }

      if (currentStyles.length) {
        properties.push(objectProperty(keyPath.node, objectExpression(currentStyles)));
      }

      return true;
    };

    const tryComposesValue = () => {
      if (!(key === 'composes' && valuePath.isArrayExpression())) {
        return false;
      }

      if (postfix !== '') {
        throw new Error('TADDY: composes could be used only on top level');
      }

      for (const elemPath of valuePath.get('elements')) {
        this.process(elemPath);
      }

      return false;
    };

    const tryEvaluateValue = () => {
      if (!this.config.evaluate) return false;
      const {
        value
      } = evaluate(valuePath);
      if (!value) return false; // console.log('evaluate', {value});

      properties.push(...this.stylesToNode(key, value, {
        postfix
      }));
      this.optimizationPaths.add(path);
      return true;
    };

    const STOP = Symbol('STOP');

    const getTSValue = tsValuePath => {
      var _this$tsProject;

      const tsType = (_this$tsProject = this.tsProject) === null || _this$tsProject === void 0 ? void 0 : _this$tsProject.getTypeAtPos(this.config.filename, this.config.code, tsValuePath.node.start || 0, tsValuePath.node.end || 0);
      if (!tsType) return; // console.log('TYPE', tsType.getType().getText());

      return parseValue(tsType.getType());
    };

    const getComposesTSValue = path => {
      const elements = [];
      elements[STATIC_KEY] = true;

      for (let elementPath of path.get('elements')) {
        if (!elementPath) return;
        const value = getTSValue(elementPath);
        if (value === STOP) return;
        elements[STATIC_KEY] = elements[STATIC_KEY] && isStaticValue(value);
        elements.push(value);
      }

      return elements;
    };

    const tryTypescriptValue = () => {
      if (!this.config.typescript) return false;
      let tsValue;

      if (key === 'composes' && valuePath.isArrayExpression()) {
        tsValue = getComposesTSValue(valuePath);
      } else {
        tsValue = getTSValue(valuePath);
      }

      if (!tsValue) return false; // console.log({tsValue});

      if (isStaticValue(tsValue)) {
        properties.push(...this.stylesToNode(key, tsValue, {
          postfix
        }));
        this.optimizationPaths.add(path);
        return true;
      } // TODO: improve composes handling


      if (key === 'composes') return false;
      const variants = [].concat(tsValue);
      let isCompilable = true; // pre-generate classes for variants

      variants.forEach(v => {
        if (v instanceof DynamicType || !isStaticValue(v)) {
          isCompilable = false;
          return;
        }

        $css({
          [key]: v
        }, {
          postfix
        });
      });
      if (!isCompilable) return false;
      const propKey = getHashedName(key, {
        postfix
      });
      properties.push(objectProperty(identifier(propKey), callExpression(identifier('css.h'), [valuePath.node])));
      return true;
    };

    const tryCSSVariableValue = () => {
      if (!this.config.CSSVariableFallback) return false; // think about dynamic nested for mixins
      // in case of nested dynamic values in mixin we can't fallback them as custom properties

      if (this.mixin || key === 'composes' || key[0] === ':' || /[\W]/.test(key)) {
        return false;
      }

      const value = getHashedName(key, {
        postfix
      });
      const dynamicValue = `--${value}`;
      this.variables.push(objectProperty(stringLiteral(dynamicValue), valuePath.node));
      properties.push(...this.stylesToNode(key, `var(${dynamicValue})`, {
        postfix
      }));
      return true;
    };

    if (tryLiteralValue() || tryObjectExpressionValue() || tryComposesValue() || tryEvaluateValue() || tryTypescriptValue() || tryCSSVariableValue()) {
      return;
    }

    properties.push(path.node);
  }

  processObjectProperty(path, {
    postfix = '',
    properties
  }) {
    const keyPath = path.get('key');
    const key = getObjectPropertyKey(keyPath);

    const tryEvaluateKey = () => {
      if (!this.config.evaluate) return false;
      const {
        value
      } = evaluate(keyPath);
      if (!(value && typeof value === 'string')) return false;
      keyPath.replaceWith(stringLiteral(value));
      this.processPropertyValue(value, path, {
        postfix,
        properties
      });
      return true;
    };

    if (key === 'className' || key === 'style' || key === null) {
      if (tryEvaluateKey()) {
        return;
      }

      properties.push(path.node);
      return;
    }

    this.processPropertyValue(key, path, {
      postfix,
      properties
    });
  }

  processSpreadElement(path, {
    postfix = '',
    properties
  }) {
    const tryEvaluateSpreadElement = () => {
      if (!this.config.evaluate) return false;
      const {
        value
      } = evaluate(path.get('argument'));
      if (!value) return false;
      const {
        className
      } = $css(value, {
        postfix
      });
      properties.push(...this.classNamesToNode(className));
      this.optimizationPaths.add(path);
      return true;
    };

    const tryTypescriptSpreadElement = () => {
      var _this$tsProject2;

      if (!this.config.typescript) return false;
      const {
        argument
      } = path.node;
      const tsType = (_this$tsProject2 = this.tsProject) === null || _this$tsProject2 === void 0 ? void 0 : _this$tsProject2.getTypeAtPos(this.config.filename, this.config.code, argument.start || 0, argument.end || 0);
      if (!tsType) return false;
      const styles = parseValue(tsType.getType());
      if (!isStaticValue(styles)) return false;
      const {
        className
      } = $css(styles, {
        postfix
      });
      properties.push(...this.classNamesToNode(className));
      this.optimizationPaths.add(path);
      return true;
    };

    if (tryEvaluateSpreadElement() || tryTypescriptSpreadElement()) {
      return;
    }

    this.process(path.get('argument'));
    properties.push(path.node);
  }

  processLogicalExpression(path, options) {
    this.process(path.get('left'), options);
    this.process(path.get('right'), options);
  }

  processObjectExpression(path, {
    postfix = ''
  }) {
    const properties = [];

    for (const propPath of path.get('properties')) {
      if (propPath.isObjectProperty()) {
        this.processObjectProperty(propPath, {
          postfix,
          properties
        });
        continue;
      }

      if (propPath.isSpreadElement()) {
        this.processSpreadElement(propPath, {
          postfix,
          properties
        });
        continue;
      }

      properties.push(propPath.node);
    }

    path.node.properties = mergeObjectProperties(properties);
  }

  process(path, options = {}) {
    if (path.isLogicalExpression()) {
      this.processLogicalExpression(path, options);
      return;
    }

    if (path.isObjectExpression()) {
      this.processObjectExpression(path, options);
      return;
    }
  }

  run(callPath, {
    mixin = false
  } = {}) {
    const args = callPath.get('arguments');
    const argPath = args[0];
    const composes = [];
    let isObject = true;

    for (let x of args) {
      isObject = isObject && x.isObjectExpression();
      composes.push(x.node);
    }

    if (isObject) {
      argPath.replaceWith(objectExpression(mergeObjects(composes)));
    } else {
      argPath.replaceWith(objectExpression([objectProperty(identifier('composes'), arrayExpression(composes))]));
    }

    callPath.node.arguments = [argPath.node];
    const path = argPath; // supports only object expressions at the moment

    if (!path.isObjectExpression()) {
      return null;
    }

    this.prepare({
      mixin
    });
    this.process(path);

    if (this.variables.length > 0) {
      path.node.properties.push(objectProperty(identifier(VARS_KEY), objectExpression(this.variables)));
    }

    const isStatic = !this.mixin && path.node.properties.every(x => {
      if (!isObjectProperty(x)) return false;
      if (!('name' in x.key)) return false;

      if (!(x.key.name[0] === '_' || x.key.name === 'className' || x.key.name === 'style' || x.key.name === VARS_KEY)) {
        return false;
      }

      return true;
    });

    if (isStatic) {
      optimizeStaticStyles(path);
    }

    return {
      isStatic,
      optimizationPaths: this.optimizationPaths
    };
  }

}

let LAST_INDEX = 0;
let STYLES = [];

function getStylesState() {
  const {
    rules
  } = ruleInjector.styleSheet;
  const newRules = rules.slice(LAST_INDEX);
  LAST_INDEX = rules.length;
  const added = [...newRules].map(rule => rule.cssText || '');
  STYLES.push(...added);
  return {
    added,
    total: STYLES
  };
}

function resolveFilepath(getter, ...params) {
  if (typeof getter !== 'string') {
    // @ts-expect-error
    return getter(...params);
  }

  return path.join(getRootDir(), getter);
}

function getEnv(babel) {
  try {
    return babel.env();
  } catch (e) {// console.log('error', e);
  }

  const DEFAULT_ENV = 'production';

  if (!(typeof process && process.env)) {
    return DEFAULT_ENV;
  }

  return process.env.BABEL_ENV || process.env.NODE_ENV || DEFAULT_ENV;
}

function getMtime(filepath) {
  return Number(fs.statSync(filepath).mtime);
}

const filesMap = /*#__PURE__*/new Map();
const contentMap = /*#__PURE__*/new Map();

function readFileSync(filepath) {
  let data = '';

  try {
    const mtime = getMtime(filepath);

    if (filesMap.has(mtime)) {
      return filesMap.get(mtime);
    }

    data = fs.readFileSync(filepath).toString();
    filesMap.set(mtime, data);
    contentMap.set(data, mtime);
  } catch (e) {}

  return data;
}

function appendFile(filepath, content, append) {
  const code = content + append;

  if (contentMap.has(code)) {
    return;
  }

  fs.appendFile(filepath, append, error => {
    if (error) {
      console.error(error);
    }
  });
}

function output({
  env,
  state,
  config: {
    // getCSSFilename = (content) => `atoms-${stringHash(content)}.css`,
    cssFilename = () => `taddy.css`,
    cssFilepath = (filename = 'taddy.css') => path.join(getCacheDir(), filename)
  } = {}
}) {
  const {
    added
  } = getStylesState();
  const filename = resolveFilepath(cssFilename);
  const filepath = resolveFilepath(cssFilepath, filename);
  const stylesData = readFileSync(filepath); // TODO: improve filter performance

  const diffStyles = added.filter(x => !stylesData.includes(x)).join('');
  appendFile(filepath, stylesData, diffStyles);
}

function isPathRemoved(path) {
  do {
    if (path.removed) return true;
  } while (path = path.parentPath);

  return false;
}

const createProcessors = (config, options) => {
  const mixinsQueue = [];
  const proceedPaths = [];
  const processor = new Processor({
    config
  });
  const handlers = {
    css: (path, {
      mixin = false
    } = {}) => {
      if (path.isMemberExpression() && 'name' in path.node.property && path.node.property.name === 'mixin') {
        mixinsQueue.push(path.parentPath);
        return;
      }

      if (!path.isCallExpression()) return;
      const result = processor.run(path, {
        mixin
      });
      if (!result) return;
      const {
        isStatic,
        optimizationPaths
      } = result;
      proceedPaths.push({
        isStatic,
        optimizationPaths
      }); // path.node.arguments = [argPath.node];
    },
    mixin: path => handlers.css(path, {
      mixin: true
    })
  };
  return {
    handlers,

    finish() {
      let isStatic = false;

      if (config.optimizeBindings) {
        isStatic = true;

        for (let x of proceedPaths) {
          isStatic = isStatic && x.isStatic;

          for (let path of x.optimizationPaths) {
            optimizeBindings(path, {
              env: options.env
            });
          }
        }
      }

      for (let path of mixinsQueue) {
        if (config.optimizeBindings && isPathRemoved(path)) {
          // the path binding should be already optimized
          continue;
        }

        isStatic = false;
        handlers.css(path, {
          mixin: true
        });
      }

      return {
        isStatic
      };
    }

  };
};

function mapCompileOptions({
  filename,
  code,
  evaluate = true,
  typescript = /\.tsx?$/.test(filename),
  unstable_CSSVariableFallback = true,
  unstable_optimizeBindings = true
}) {
  return {
    filename,
    code,
    evaluate,
    typescript,
    CSSVariableFallback: unstable_CSSVariableFallback,
    optimizeBindings: unstable_optimizeBindings
  };
}

function macro({
  references,
  babel,
  state,
  config: _config = {}
}) {
  const {
    env = getEnv(babel)
  } = _config,
        config = _objectWithoutPropertiesLoose(_config, ["env"]);

  if (isTaddyEvaluation(state)) {
    return;
  }

  const program = state.file.path;
  const code = state.file.code;
  const {
    filename
  } = state;
  const {
    handlers,
    finish
  } = createProcessors(mapCompileOptions(_extends({}, config.compileOptions, {
    filename,
    code
  })), {
    env
  });

  for (const key in references) {
    if (!handlers[key]) continue;
    references[key].forEach(path => handlers[key](path.parentPath));
  }

  let importPath;
  program.traverse({
    ImportDeclaration(p) {
      const {
        node
      } = p;

      if (!(node.source.value === PACKAGE_NAME || node.source.value === MACRO_NAME)) {
        return;
      }

      importPath = p;
      importPath.node.source = stringLiteral('taddy');
      p.stop();
    }

  });
  const {
    isStatic
  } = finish();

  if (isStatic && importPath) {
    importPath.node.source = stringLiteral('@taddy/core');
  }

  output({
    env,
    state,
    config: config.outputOptions
  });
  return {
    keepImports: true
  };
}

function getImportNames(path) {
  const localName = path.node.local.name;

  if (path.isImportSpecifier()) {
    return {
      localName,
      importedName: path.node.imported.name
    };
  }

  if (path.isImportDefaultSpecifier()) {
    return {
      localName,
      importedName: 'default'
    };
  }

  return {
    localName,
    importedName: 'namespace'
  };
}

function findReferences(paths) {
  const references = {};

  for (let path of paths) {
    const {
      localName,
      importedName
    } = getImportNames(path);
    const binding = path.scope.getBinding(localName);

    if (binding === null || binding === void 0 ? void 0 : binding.referencePaths) {
      references[importedName] = binding.referencePaths;
    }
  }

  return references;
}

function plugin(babel, options) {
  const env = getEnv(babel);
  return {
    name: 'taddy',

    pre() {
      this.references = {};
    },

    visitor: {
      // TODO: support require expression
      ImportDeclaration(path, state) {
        if (isTaddyEvaluation(state)) {
          return;
        }

        if (path.node.source.value !== PACKAGE_NAME) {
          return;
        }

        Object.assign(this.references, findReferences(path.get('specifiers')));

        if (!(this.references.css || this.references.mixin)) {
          return;
        }

        macro({
          references: this.references,
          babel,
          state,
          config: _extends({
            env
          }, options)
        });
      }

    }
  };
}

export default plugin;
export { macro };
//# sourceMappingURL=babel-plugin.esm.js.map
